/**
Developed for Hobbyiest use by Tomas Diaz-Wahl for Kalliot LLC.

Copyright © 2022 Kalliot LLC

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
conditions are met:
1. Redistributions of source code and documentation must retain the above copyright notice, this list of conditions and the 
following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
disclaimer in the documentation and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products
derived from this software without specific prior written permission.

THIS SOFTWARE AND DOCUMENTATION ARE PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/

/*
 * This project uses two third-party libraries.
 * INA260 and HX711 libraries can be found at:
 * <https://learn.adafruit.com/adafruit-ina260-current-voltage-power-sensor-breakout/arduino>
 * <https://github.com/bogde/HX711>
 */
 
#include <Adafruit_INA260.h>
#include <Servo.h>
#include "HX711.h"


Adafruit_INA260 ina260 = Adafruit_INA260();
Servo servo;
HX711 scale;

/*
 * Debug class is designed to output info to the serial monitor, (found under the tools menu in the Arduino IDE),
 * This is designed to make sure that all boards and electronics are working as designed.
 * TO ACTIVATE DEBUG CLASS MAKE SURE debugOn IS SET TO TRUE.
 */

class Debug {
  public:
    static const bool debugOn = false; //turn debug on and off
    void printMsg(int x){
      if (debugOn){
        Serial.print(x);
        Serial.flush();
      }
    }

    void printMsg(String x){
      if (debugOn){
        Serial.print(x);
        Serial.flush();
      }
    }

    void printLine(int x){
      if (debugOn){
        Serial.println(x);
        Serial.flush();
      }
    }

    void printLine(String x){
      if (debugOn){
        Serial.println(x);
        Serial.flush();
      }
    }

    void setUpSerial(){
      if (debugOn){
        Serial.begin(9600);
        // Wait until serial port is opened
        while (!Serial) { delay(10); }
      }
    }
};//Debug

/*
 * Queue class is designed to make a basic queue for our current measurements. 
 * There is no need to modify this class.
 */

class Queue{
  private: 
    const static int maxQueueSize = 8;
    float arr [maxQueueSize];
    int queuePos;
    
  public: 
    Queue(){
        for (int i = 0; i < maxQueueSize; i++)
        {
          arr[i] = 0;
        }
        queuePos = maxQueueSize - 1;
    }// Queue constructor
    
    int enqueue (float val){
        queuePos += 1;
        if (queuePos >= maxQueueSize){
            queuePos = 0;
        }
        arr[queuePos] = val;
    }
    
    float getAvg (){
          float total = 0;
          float avg;

          for (int i = 0; i < maxQueueSize; i++){
            total += arr [i]; 
          }
          avg = total / maxQueueSize;
          return avg;
    }
};// class Queue

/*
 * Feeder class allows for motion of the feeder lid as well as implementing a check 
 * to see if there is an obstruction affecting lid motion.
 * You will need to change several variable in the private data section of this class.
 * You MUST change: calibrationFactor, minCatWeight, maxCatWeight
 * You MAY need to change: maxCurrent, lidMaxPosition, any pin designations you wish to change.
 */

class Feeder {
    public:
        
        const int moveAmount = 1; // How much the servo should move per command

        /*
         * Set up stores an known initial position in the microcontroller memory.
         * Tells the microcontroller which pin the servo is on.
         * Activates and sets the scale calibration factor (see SparkFun_HX711_Calibration sketch)
         */

        void setUp () {
            servo.write (initialPos);
            servo.attach(servoPin);
            delay (1000);
            scale.begin(LOADCELL_DOUT_PIN, LOADCELL_SCK_PIN);
            scale.set_scale(calibrationFactor); //This value is obtained by using the SparkFun_HX711_Calibration sketch
        }

        /*
         * reset should be called when powering up the feeder but can be called manually as well.
         * reset tares the scale as well as moves the servo to a known position and then slowly moves the servo to position 0.
         */
        
        void reset () {
            scale.tare(); //Assuming there is no weight on the scale at start up, reset the scale to 0
            servo.write (initialPos);
            delay (500);
            for (int i = initialPos; i > lidMinPosition; i--){
              servo.write (i);
              delay (servoDelay); 
            }
            pos = lidMinPosition;
            numberOfRetries = 0;
        }
        
        bool IsCatOn ()
        {
          float weight = scale.get_units();
          return ((weight >= minCatWeight) && (weight <= maxCatWeight));
        }

        void moveServo (int amount){
            pos += amount;
            servo.write(pos);
        }

        void raiseLid (int amount){
            moveServo(amount);
        }
        
        void lowerLid (int amount){
            moveServo (-amount);
        }

        /*
         * Open and close lid are nearly identical in function (haha! get it? they're actually methods)
         * These methods are designed to use the Adafruit INA260 current sensor to detect the current during operation.
         * If the current draw from the servo is too high, it may indicate an obstruction. 
         * The implementation of this is a little complex so I recommend not messing with these values.
         */

        float openLid () {
          while ((pos < lidMaxPosition) && (numberOfRetries < maxRetriesAllowed))
          {
            raiseLid(moveAmount);
            float current = ina260.readCurrent();
            myQueue.enqueue(current);
            if (myQueue.getAvg() >= maxCurrent){
                    numberOfRetries += 1;
                    lowerLid(5);
                    delay(1000);//Wait before lid open retry
                    debug.printLine("Max Current Detected, Closing Lid Before Retry");
                    if(numberOfRetries > maxRetriesAllowed){
                      exit (-1);//Lid is blocked
                    }
                    closeLid();
                    return;
            }
            delay(servoDelay);
          } // end while loop
          numberOfRetries = 0;//reset counter
        } // openLid

        void closeLid (){
         while (pos > lidMinPosition)
          {
            lowerLid(moveAmount);
            float current = ina260.readCurrent();
            myQueue.enqueue(current);
            if (myQueue.getAvg() >= maxCurrent){
                    numberOfRetries += 1;
                    raiseLid(5);
                    delay(1000);
                    debug.printLine("Max Current Detected, Opening Lid Before Retry");
                    if(numberOfRetries > maxRetriesAllowed){
                      exit (-1);//Lid is blocked
                    }
                    openLid();
                    return;
            }
            delay(servoDelay);
          } // end while loop
          numberOfRetries = 0; //reset retry counter
        } // CloseLid
        
    private:
        const float calibrationFactor = -9660.0; //This value is obtained using the SparkFun_HX711_Calibration sketch.
        const int LOADCELL_DOUT_PIN  = 10;
        const int LOADCELL_SCK_PIN = 11;
        const int servoDelay = 40; // Speed of servo. (change not recommended)
        const int initialPos = 20; // initial position of servo when feeder is reset. Be careful!; servo will move at max speed to this position. 
        const int lidMaxPosition = 70; // Maximum position of servo may vary depending on tolerances.
        const int lidMinPosition = 0; 
        const int servoPin = 9;
        const float minCatWeight = 11.0; // lbs. Set the minimum activation weight of your animal.
        const float maxCatWeight = 1000.0; // lbs. Set the maximum weight of animal. (if you have a bigger dog or cat you dont want getting into the feeder)
        const float maxCurrent = 250.00;  // Maximum current that is allowed under normal feed operation. Triggers overcurrent protection mechanism. 
        const int maxRetriesAllowed = 10; // Maximum number of lid movement commands during overcurrent protection before lid operation ceases. 
        const int numberOfMeasurements = 5; // Number of current measurements to be averaged.
        int pos;
        int numberOfRetries;
        Queue myQueue;
        Debug debug;
        
}; // Feeder



Feeder feeder;
Debug debug;

void setup() {
  delay(1000); //wait for boot up
  // DEBUG BEGIN
  debug.setUpSerial(); // Enables serial port communication. (Make sure debugOn set to true in debug class for USB serial communication) 

  debug.printLine("Adafruit INA260 Test");

  if (!ina260.begin()) {
    debug.printLine("Couldn't find INA260 chip");
    exit (-1);
  }
  debug.printLine("Found INA260 chip");
  debug.printLine("Configuring Ina");
  debug.printMsg("INA260_TIME_8_244_ms = ");
  debug.printLine(INA260_TIME_8_244_ms);
  
  ina260.setCurrentConversionTime(INA260_TIME_8_244_ms); // Changes the measurement interval. This makes current readings less spikey.
  
  debug.printMsg("INA260_COUNT_4 = ");
  debug.printLine(INA260_COUNT_4);
  
  ina260.setAveragingCount(INA260_COUNT_4); // Changes the number of measurements averaged by the INA260 sensor. This will also make readings less spikey.
  INA260_ConversionTime convTime = ina260.getCurrentConversionTime();  // Verifying Ina Configuration
  INA260_AveragingCount avgCount = ina260.getAveragingCount();   // Verifying Ina Configuration
  
  if (convTime != INA260_TIME_8_244_ms){
    exit(-1);
  }
  if (avgCount != INA260_COUNT_4){
    exit(-1);
  }
  
  debug.printLine("Configured Ina");

  // DEBUG END

   feeder.setUp(); // initialize feeder.
   feeder.reset(); // set feeder lid to position 0 and tare scale.

  delay(2000);// During construction of the feeder, this delay can be set much higher to align the servo gear with the door gear. (avoid pinchy fingies)
}


void loop() {

  debug.printLine("I'm in the loop");

  // The following code was used during construction for debug purposes
  //  feeder.openLid();
  //  feeder.closeLid();
  //debug.printLine("Program Exit");
  //
  //  exit(0);
  //END DEBUG
  
  //main code for feeder operation
  if (feeder.IsCatOn()) {
    feeder.openLid();
  }
  else{ // Cat is not on the feeder
    feeder.closeLid();
  }

  delay(1000);
}